#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
(c) 2014 - Ronan Delacroix
FTP utils
:author: Ronan Delacroix
"""
import sys
try:
    import logging
    import pyftpdlib.servers
    import pyftpdlib.handlers
    from . import text

    try:
        from pyftpdlib.handlers import TLS_FTPHandler as SFTPHandler
    except ImportError:
        from pyftpdlib.handlers import FTPHandler as SFTPHandler  # If that case happens, install pyopenssl


    class FTPEventLogger:

        def on_connect(self):
            self.log("Connection received.")

        def on_disconnect(self):
            self.log("Disconnection.")

        def log(self, msg, logfun=None, error=False):
            raise Exception('Sub classes should override this function.')

        def on_login(self, username):
            self.log("User %s logged in." % username)

        def on_login_failed(self, username, password):
            self.log("Login failed with credentials %s/%s." % (username, password))

        def on_logout(self, username):
            self.log("User %s logged out." % username)

        def on_file_sent(self, filepath):
            filepath = text.convert_to_unicode(filepath)
            self.log(u"File %s has been successfully sent (User %s)." % (filepath, self.username))

        def on_file_received(self, filepath):
            filepath = text.convert_to_unicode(filepath)
            self.log(u"User %s has uploaded a new file : %s" % (self.username, filepath))

        def on_incomplete_file_sent(self, filepath):
            filepath = text.convert_to_unicode(filepath)
            self.log("""File %s has been Incompletely sent by user %s...
            Waiting for the user to resume his download.""" % (filepath, self.username), error=True)

        def on_incomplete_file_received(self, filepath):
            filepath = text.convert_to_unicode(filepath)
            self.log("""A new file %s has been uploaded but is incomplete by user %s...
            Waiting for the user to resume his upload.""" % (filepath, self.username), error=True)


    class FTPHandler(FTPEventLogger, pyftpdlib.handlers.FTPHandler):

        def log(self, msg, logfun=None, error=False):
            if error:
                logging.error("[FTP] %s" % msg)
            else:
                logging.info("[FTP] %s" % msg)


    class SecureFTPHandler(FTPEventLogger, SFTPHandler):

        def log(self, msg, logfun=None, error=False):
            if error:
                logging.error("[FTPS] %s" % msg)
            else:
                logging.info("[FTPS] %s" % msg)


    class DummyDictFTPAuthorizer(pyftpdlib.handlers.DummyAuthorizer):
        """
        Dummy Dict FTP Authorizer class.
        Provide authentication through FTP for users stored in a dict.

        About permissions:

            Read permissions:
             - "e" = change directory (CWD command)
             - "l" = list files (LIST, NLST, STAT, MLSD, MLST, SIZE, MDTM commands)
             - "r" = retrieve file from the server (RETR command)

            Write permissions:
             - "a" = append data to an existing file (APPE command)
             - "d" = delete file or directory (DELE, RMD commands)
             - "f" = rename file or directory (RNFR, RNTO commands)
             - "m" = create directory (MKD command)
             - "w" = store a file to the server (STOR, STOU commands)
        """

        def __init__(self, users):
            """
            Constructor
            """
            super(DummyDictFTPAuthorizer, self).__init__()
            for username, user in users.items():
                self.add_user(username,
                              user['password'],
                              user['homedir'],
                              perm=user.get('perm', self.read_perms),
                              msg_login="Hi %s, you're welcome here." % user.get('name', username),
                              msg_quit="Bye %s, hoping you get back soon!" % user.get('name', username)
                )
            self.custom_users = users


    def create_server(handler, users, listen_to="", port=21, data_port_range='5500-5700', name="Ronan Python FTP Server", masquerade_ip=None, max_connection=500, max_connection_per_ip=10):
        """
        Runs the FTP Server
        """

        try:
            start, stop = data_port_range.split('-')
            start = int(start)
            stop = int(stop)
        except ValueError:
            raise Exception('Invalid value for data ports')
        else:
            data_port_range = range(start, stop + 1)

        handler.authorizer = DummyDictFTPAuthorizer(users=users)

        handler.banner = "%s. (Advice : Please use UTF-8 encoding and always use Binary mode)" % name
        handler.passive_ports = data_port_range

        if masquerade_ip:
            handler.masquerade_address = masquerade_ip

        #logging.getLogger('pyftpdlib').disabled = True
        pyftpdlib.log.logger = logging.getLogger()  # Replace pyftpd logger by default logger

        # Instantiate FTP server class and listen to 0.0.0.0:21 or whatever is written in the config
        address = (listen_to, port)
        server = pyftpdlib.servers.FTPServer(address, handler)

        # set a limit for connections
        server.max_cons = max_connection
        server.max_cons_per_ip = max_connection_per_ip

        return server


    def create_ftp_server(users, listen_to="", port=21, data_port_range='5500-5700',
                          name="FTP Server", masquerade_ip=None, max_connection=500, max_connection_per_ip=10):
        """
            FTP Server implements normal FTP mode.
        """
        handler = FTPHandler
        return create_server(handler, users, listen_to=listen_to, port=port, data_port_range=data_port_range,
                             name=name, masquerade_ip=masquerade_ip, max_connection=max_connection,
                             max_connection_per_ip=max_connection_per_ip)


    def create_secure_ftp_server(users, certificate, listen_to="", port=990, data_port_range='5700-5900',
                                 name="FTP Server", masquerade_ip=None, max_connection=500, max_connection_per_ip=10):
        """
            FTP Server implements FTPS (FTP over TLS/SSL) mode.
              Note: Connect from client using "FTP over TLS/SSL explicit mode".
        """
        handler = SecureFTPHandler
        handler.certfile = certificate
        return create_server(handler, users, listen_to=listen_to, port=port, data_port_range=data_port_range,
                             name=name, masquerade_ip=masquerade_ip, max_connection=max_connection,
                             max_connection_per_ip=max_connection_per_ip)

except ImportError:
    print("Impossible to import FTP Server helpers from tbx.ftp module. Requires pyftpdlib.", file=sys.stderr)

try:
    """
    Following is only for client FTP using FTPUtil library.
    """
    import ftplib


    class FTPSession(ftplib.FTP):
        """
        This class is required to connect to a different port with FTPUtil lib.
        """
        def __init__(self, url, user, password, port=21, passive=None, timeout=None):
            if timeout:
                ftplib.FTP.__init__(self, timeout=timeout)
            else:
                ftplib.FTP.__init__(self)
            self.connect(url, port)
            self.login(user, password)
            if passive is not None:
                self.set_pasv(passive)

        """
        Usage :
            import ftputil
            host = ftputil.FTPHost('my.server.url.or.ip', user='root', password='hello', 2121, session_factory=FTPSession)
            names = host.listdir(host.curdir)
            print(names)
        """

except ImportError:
    print("Impossible to import FTP helpers from tbx.ftp module. Requires ftplib.", file=sys.stderr)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
(c) 2013 - Ronan Delacroix
Bytes Utils
:author: Ronan Delacroix

Generic byte functions.
Should work in Python 3 and 2
"""
import socket
import struct
import six
import uuid
import binascii
from . import network


def bytes_to_int(byte_array, big_endian=True, signed=False):
    """
    Converts a byte array to an integer.
    """
    if six.PY3:
        order = 'little'
        if big_endian:
            order = 'big'
        return int.from_bytes(byte_array, byteorder=order, signed=signed)
    else:
        length = len(byte_array)
        if length == 1:
            code = 'B'
        elif length == 2:
            code = 'H'
        elif length == 4:
            code = 'L'
        elif length == 8:
            code = 'Q'
        else:
            raise Exception("bytes_to_int : length of byte_array should be 1, 2, 4, or 8")
        if big_endian:
            code = '>'+code
        else:
            code = '<'+code

        if signed:
            code = code.lower()
        return struct.unpack(code, byte_array)[0]


def bytes_to_uuid(byte_array):
    """
    Converts a byte array (should be 16 bytes length) to a uuid object.
    :param byte_array: the byte array to convert
    :return: uuid object
    """
    return uuid.UUID(bytes=bytes(byte_array))


def uuid_to_bytes(id):
    """
    Convert a uuid to a 16-byte string (containing the six integer fields in big-endian byte order)
    :param id: uuid string or object
    :return: a byte array
    """
    return uuid.UUID(id).bytes


def bytes_to_uuid_list(byte_array):
    """
    Converts a byte array to a list of uuids. Cuts the byte array by packets of 16 bytes and parse each as uuid.
    :param byte_array: a byte array of length n*16
    :return: a list of uuid objects
    """
    result = []
    for i in range(0, len(byte_array)//16):
        result.append(uuid.UUID(bytes=bytes(byte_array[i*16:i*16+16])))
    return result


def batch(byte_array, funcs):
    """
    Converts a batch to a list of values.
    :param byte_array: a byte array of length n*item_length + 8
    :return: a list of uuid objects
    """
    result = []
    length = bytes_to_int(byte_array[0:4])
    item_size = bytes_to_int(byte_array[4:8])
    for i in range(0, length):
        chunk = byte_array[8+i*item_size:8+(i+1)*item_size]
        for f in funcs:
            f(chunk)
    return result


def bytes_to_text(byte_array, encoding='UTF-8'):
    """
    Decode a byte array to a string following the given encoding.
    :param byte_array: Byte array to decode.
    :param encoding: String encoding (default UTF-8)
    :return: a decoded string
    """
    return bytes(byte_array).decode(encoding)


def text_to_bytes(text, encoding='UTF-8', size=None):
    """
    Encode some text or string to a byte array
    :param text: text to encode to bytes
    :param encoding: optional encoding of the passed string. default to utf-8.
    :param size: optional, if given the text will be padded with 0x00 to the right size
    :return: a bytes object
    """
    res = str(text).encode(encoding)
    if size:
        res = res.rjust(size, b'\x00')
    return res


def bytes_to_hex(byte_array):
    """
    Converts byte to hex string...
    :param byte_array: byte array to return as hex
    :return: an ascii string, hex representation of the byte array
    """
    return binascii.hexlify(bytes(byte_array)).decode('ascii')


def int_to_bytes(val, bit=32, signed=False, big_endian=True):
    """
    Converts an int to a byte array (bytes).
    :param val: value to encode
    :param bit: bit length of the integer to encode
    :param signed: encode as unsigned int if false
    :param big_endian: encode with big or little endian
    :return:
    """
    val = int(val) #ensure it is an int

    if six.PY3:
        order = 'little'
        if big_endian:
            order = 'big'
        return val.to_bytes(length=bit//8, byteorder=order, signed=signed)

    if bit == 8:
        code = 'B'
    elif bit == 16:
        code = 'H'
    elif bit == 32:
        code = 'I'
    elif bit == 64:
        code = 'Q'
    else:
        raise Exception("int_to_bytes : size parameter value should be 8, 16, 32, or 64")

    if big_endian:
        code = '>'+code
    else:
        code = '<'+code

    if signed or val < 0:
        code = code.lower()

    return struct.pack(code, val)


def ip_to_bytes(ip_str, big_endian=True):
    """
    Converts an IP given as a string to a byte sequence
    """
    if big_endian:
        code = '>L'
    else:
        code = '<L'
    return bytes(struct.unpack(code, socket.inet_aton(ip_str))[0])


def decode_ber(ber):
    """
    Decodes a ber length byte array into an integer
    return: (length, bytes_read) - a tuple of values
    """
    ber = bytearray(ber)
    length = ber[0]
    bytes_read = 1
    if length > 127:
        bytes_read += length & 127 # Strip off the high bit
        length = 0
        for i in range(1, bytes_read):
            length += ber[i] << (8 * (bytes_read - i - 1))
    return length, bytes_read


def ber_from_socket(s):
    """
    Reads a ber length from a socket.
    Reads first byte to read the ber length first, then read the actual data.
    return: (length, bytes_read, data) - a tuple of values
    """
    if isinstance(s, network.SocketClient):
        recv = s.receive
    else:
        recv = s.recv
    data = recv(1)

    length = data[0]
    bytes_read = 1
    if length > 127:
        bytes_read += length & 127 # Strip off the high bit
        data += recv(bytes_read - 1)
        length = 0
        for i in range(1, bytes_read):
            length += data[i] << (8 * (bytes_read - i - 1))
    return length, bytes_read, data


def encode_ber(value, ber_length=0):
    """
    Encodes an integer to ber length
    You can set the ber_length manually.
    return: bitearray object
    """
    if not ber_length:
        if value < 127: # 1 byte case
            return bytearray([value])
        ber_length = 1
        while value >= pow(2, 8*ber_length):
            ber_length += 1
        ber_length += 1


    ber = bytearray(ber_length)
    ber[0] = 127 + ber_length #ber length byte
    for i in range(1, ber_length):
        ber[i] = (value >> (8 * (ber_length - 1 - i))) & 255
    return ber




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
(c) 2013 - Ronan Delacroix
Templating Utils
:author: Ronan Delacroix
"""
import re
import jinja2
from functools import wraps
import uuid


def create_jinja_env(template_path):
    """
    Creates a Jinja2 environment with a specific template path.
    """
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_path),
        block_start_string='{%',
        block_end_string='%}',
        variable_start_string='${',
        variable_end_string='}',
        comment_start_string='{#',
        comment_end_string='#}',
        line_statement_prefix=None,
        line_comment_prefix=None,
        trim_blocks=True,
        lstrip_blocks=True,
        newline_sequence='\n'
    )
    jinja_env.filters['regexreplace'] = regex_replace
    jinja_env.globals.update(uuidgen=uuidgen)
    return jinja_env


def render_template(env, filename, values=None):
    """
    Render a jinja template
    """
    if not values:
        values = {}
    tmpl = env.get_template(filename)
    return tmpl.render(values)


def template(filename):
    """
    Decorator
    """
    def method_wrapper(method):

        @wraps(method)
        def jinja_wrapper(*args, **kwargs):
            ret = method(*args, **kwargs)
            return render_template(filename, ret)

        return jinja_wrapper

    return method_wrapper


def regex_replace(s, find, replace):
    """
    A non-optimal implementation of a regex filter (used as jinja filter)
    """
    return re.sub(find, replace, s)


def uuidgen():
    """
    Jinja filter to create random UUIDs
    """
    return uuid.uuid4()
